/*
 * *********************************************************
 * Copyright (c) 2009 - 2014, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: Jul 13, 2014
 * Author(s): Lukas Magel
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.autoreferee;

import edu.dhbw.mannheim.tigers.sumatra.model.data.frames.AIInfoFrame;
import edu.dhbw.mannheim.tigers.sumatra.model.data.modules.ai.EGameState;


/**
 * Base class for all referee case implementation which are stateless
 * and only evaluate their case based on one frame at
 * a time. The <code>ARefereeCase</code> keeps track of the current
 * game state and only forwards the {@link AIInfoFrame} to a child class
 * via {@link ARefereeCase#checkCase(AIInfoFrame)} if the game is running.
 * 
 * @author Lukas Magel
 */
public abstract class ARefereeCase implements IRefereeCase
{
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private enum ERefereeState
	{
		PAUSED,
		RUNNING,
		PENDING
	}
	
	private static final int		MAX_PENDING_TIME	= 3000;
	private static final double	CALL_FREQ			= 60.0;						// Hz
																									
	private ERefereeState			state					= ERefereeState.PAUSED;
	private int							pendingTime			= 0;							// ms
																									
																									
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * 
	 */
	public ARefereeCase()
	{
		
	}
	
	
	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public RefereeCaseMsg onNewAIInfoFrame(final AIInfoFrame frame)
	{
		return runCase(frame);
	}
	
	
	private RefereeCaseMsg runCase(final AIInfoFrame frame)
	{
		// Nothing to do
		switch (state)
		{
			case PAUSED:
				if (isGameRunning(frame))
				{
					state = ERefereeState.RUNNING;
				}
				return null;
			case PENDING:
				if (!isGameRunning(frame))
				{
					state = ERefereeState.PAUSED;
					break;
				}
				pendingTime += 1000.0 / CALL_FREQ;
				if (pendingTime >= MAX_PENDING_TIME)
				{
					state = ERefereeState.RUNNING;
				}
				return null;
			case RUNNING:
				if (!isGameRunning(frame))
				{
					state = ERefereeState.PAUSED;
					break;
				}
				RefereeCaseMsg msg = checkCase(frame);
				if (msg != null)
				{
					pendingTime = 0;
					state = ERefereeState.PENDING;
				}
				return msg;
		}
		return null;
	}
	
	
	@Override
	public void reset()
	{
		state = ERefereeState.PAUSED;
		pendingTime = 0;
	}
	
	
	private boolean isGameRunning(final AIInfoFrame frame)
	{
		return frame.getTacticalField().getGameState() == EGameState.RUNNING;
	}
	
	
	/**
	 * Called by {@link ARefereeCase} whenever a new frame has arrived and the game
	 * is running. Should be implemented by all child classes to evaluate a certain
	 * condition.
	 * 
	 * @param frame the {@link AIInfoFrame} used to evaluate the condition
	 * @return case msg if a violation has been detected
	 */
	protected abstract RefereeCaseMsg checkCase(AIInfoFrame frame);
	
	// --------------------------------------------------------------------------
	// --- getter/setter --------------------------------------------------------
	// --------------------------------------------------------------------------
	
}
